// G. Greenleaf
// K. Nguyen
// The Breakout module
module breakout (clk, reset, en, enBall, l, r, g, pauseext, ballX, ballY, paddleX, bricks, score, lives);

// i/o
input clk, reset, l, r, g, pauseext, en, enBall;
output [9:0] ballX, paddleX;
output [8:0] ballY;
output [59:0] bricks;
output [15:0] score;
output [1:0] lives;

// regs
reg [9:0] ballX, paddleX; // balls' x/y pos, paddle's x pos
reg [8:0] ballY; 
reg [59:0] bricks, newBricks; // bricks are on/off.  Color is determined by row.
reg [5:0] bricknum;  // numbering for the bricks starts in top-left corner, moving to the right per row
reg [15:0] score; // keep the user's score
reg inBrick, newLevel, newLife;  // are we in the brick; are we starting a new level; did player just die
reg [3:0] balldir, tmpDir;  // 4-bit direction fields, that say which way the ball is going
reg [1:0] lives;  // keep the player's life total
reg godMode, pauseint;  // are we cheating (key: g), are we paused (key: p)
reg [2:0] level;  // the current level that the user is on
reg lup, rup;  // signals used for the purpose of pausing... this tells whether user was pressing left or
  // right at the time of the internal reset

// wires
wire [9:0] tmpXa, tmpXb; // the new x/y coord, if we are not changing direction
wire [8:0] tmpYa, tmpYb; // the new x/y coord, if we are changing direction
wire hittingBrick;  // are we hitting the brick?
wire hitBrick;      // are we hitting the brick, part 2?
wire pause;

assign pause = pauseint | pauseext;

// figure out if we are hitting the brick: if bricknum is 0, we aren't, if it's above 60, then we are either
// hitting a paddle or the wall and so we are, otherwise we are in the normal brick range and we check the
// array to see if the brick is there
assign hitBrick = (bricknum == 0) ? 0 : 1;
assign hittingBrick = (bricknum == 0 || bricknum > 60) ? hitBrick : bricks[bricknum-1];


// the instantiations of the main modules
// moduleA: calculates bricknum based on the x/y coordinate (assuming we are not changing direction)
// moduleB: calculates the new x/y coord, if we are not changing direction
// moduleC: calculates the new x/y coord, if we are changing direction (i.e., using tmpDir)
// moduleD: uses result from B to see what direction we would change to if we are hitting a brick
inbrick moduleA(clk, tmpXa, tmpYa, paddleX, bricknum);
nextBallPos moduleB(clk, balldir, ballX, ballY, tmpXa, tmpYa);
nextBallPos moduleC(clk, tmpDir, ballX, ballY, tmpXb, tmpYb);
nextBallDir moduleD(clk, paddleX, balldir, ballX, ballY, tmpXa, tmpYa, tmpDir);


// calculate the paddle's next position
always @(posedge clk) begin
  if (en) begin
    if (reset)
      paddleX <= 320;
    else if (l && paddleX > 31 && ~pause)
      paddleX <= paddleX - 2;
    else if (r && paddleX < 609 && ~pause)
      paddleX <= paddleX + 2;
  end
end


// calculate the ball's next position, and the bricks
always @(posedge clk) begin

  // constantly be looking at the score to see if we can do some renormalization on it
  // we are 'renormalizing' to make sure we still have a valid decimal representation
  if (score[3:0] >= 10) begin
    score[3:0] <= score[3:0] - 10;
    score[7:4] <= score[7:4] + 1;
  end else if (score[7:4] >= 10) begin
    score[7:4] <= score[7:4] - 10;
    score[11:8] <= score[11:8] + 1;
  end else if (score[11:8] >= 10) begin
    score[11:8] <= score[11:8] - 10;
    score[15:12] <= score[15:12] + 1;
  end

  // turn on god mode, if requested
  if (g)
    godMode <= 1;

  // if we are no longer hitting the l or r key, then update lup and rup
  if (!l)
    lup <= 0;  
  if (!r)
    rup <= 0;

  // unpause
  if (((l & !lup) | (r & !rup)) && pauseint) begin
    pauseint <= 0;
    ballX <= paddleX;
  end

  // start over
  if (reset) begin
    godMode <= 0;
    score <= 0;
    lives <= 3;
    level <= 0;
    bricks <= 60'hFFFFFFFFFFFFFFF;
    lup <= 0;
    rup <= 0;
  end
  if (reset | newLevel) begin  // starting over or starting new level
    newLevel <= 0;
  end
  if (reset | newLevel | newLife) begin // start over, start new level, just died
    pauseint <= 1; // pause it
    
    // record whether we were going left or right
    if (l)
      lup <= 1;
    if (r)
      rup <= 1;
    newLife <= 0;
    
    // reset the ball
    if (reset)
      ballX <= 320;
    else
      ballX <= paddleX;
    ballY <= 456;
    balldir <= 0;
  end else if (~pause && enBall && lives != 0) begin  // the normal game commences!
    if (ballY >= 478) begin // hit bottom, just died
      if (~godMode)
        lives <= lives - 1;
      newLife <= 1;
    end

    if (bricks == 0) begin // means we hit the last brick, so setup the new level
      if (level == 7)
        bricks <= 60'hFFFFFFFFFFFFFFF; // since bricks is a 60-bit number, a new level is defined by a 15-digit hex number.  Cool, huh?
      else if (level == 0)
        bricks <= 60'h0C0301E0FC7FBFF;
      else if (level == 1)
        bricks <= 60'hAA955AA955AA955;
      else if (level == 2)
        bricks <= 60'hAD6B5AD6B5AD6B5;
      else if (level == 3)
        bricks <= 60'hDDD5555555555DD;
      else if (level == 4)
        bricks <= 60'hFFCFC0C0303F3FF;
      else if (level == 5)
        bricks <= 60'h9D2549D217857F5;
      else if (level == 6)
        bricks <= 60'hFFF03FCC03FFFFF;

      newLevel <= 1;
      level <= level + 1;
    end

    // if we are actually hitting a brick (and not hitting paddle or walls)      
    if ((hittingBrick & hitBrick) && (bricknum < 61)) begin
      // update score
      if (bricknum > 50)
        score <= score + 1;
      else if (bricknum > 40)
        score <= score + 2;
      else if (bricknum > 30)
        score <= score + 3;
      else if (bricknum > 20)
        score <= score + 4;
      else if (bricknum > 10)
        score <= score + 5;
      else if (bricknum > 0)
        score <= score + 6;   
        
      bricks[bricknum-1] <= 0;  // remove the brick
    end

    // if we are hitting anything that causes the ball to change direction, then change direction
    if (hittingBrick) begin    
      ballX <= tmpXb;
      ballY <= tmpYb;
      balldir <= tmpDir;
    end else begin
      ballX <= tmpXa;
      ballY <= tmpYa;
    end
  end       
end

endmodule


// module to determine if we are in a brick, and if so, which brick.  63 is bricknum for paddle.  62 for OB.
module inbrick(clk, bx, by, paddleX, bricknum);

input clk;
input [9:0] bx, paddleX;
input [8:0] by;
output [5:0] bricknum;

wire [7:0] bnum;
wire [7:0] ynum;
reg [5:0] bricknum;

// calculate the brick number (simple x/y transformation)
assign ynum = by[8:5];
assign bnum = (ynum >= 3 && ynum <= 8) ? (10 * ynum + bx[9:6] - 29) : 0;

always @(posedge clk) begin
  if ((by >= 458) && (by <= 460) && (bx > paddleX - 30) && (bx < paddleX + 30))
    bricknum <= 63; // in paddle
  else if (!(bx < 638 && bx > 2 && by <= 478 && by >= 2))
    bricknum <= 62; // OB
  else if (bnum > 60 || bnum==0) 
    bricknum <= 0;  // outside brick region
  else
    bricknum <= bnum; // inside brick region
end


endmodule



// module to calculate the new x/y coordinate, based on direction and current x/y
module nextBallPos(clk, bdir, bx, by, outx, outy);

input clk;
input [3:0] bdir;
input [9:0] bx;
input [8:0] by;
output [9:0] outx;
output [8:0] outy;

reg [9:0] outx;
reg [8:0] outy;

// direction is set up thusly:
// bdir[3] says whether we are going left (=0) or right (=1)
// bdir[1] says whether we are going 1 pixel (=1) or 2 pixels (=0) per iteration in the x direction
// bdir[2] says whether we are going up (=0) or down (=1)
// bdir[0] says whether we are going 1 pixel (=1) or 2 pixels (=0) per iteration in the y direction
always @(posedge clk) begin
  if (bdir[3]) begin
    if (bdir[1])
      outx <= bx + 1;
    else
      outx <= bx + 2;
  end else begin
    if (bdir[1])
      outx <= bx - 1;
    else
      outx <= bx - 2;
  end
      
  if (bdir[2]) begin
    if (bdir[0])
      outy <= by + 1;
    else
      outy <= by + 2;
  end else begin
    if (bdir[0])
      outy <= by - 1;
    else
      outy <= by - 2;
    
  end
end

endmodule


// calculate the direction of the next ball, based on where we are now, where we are going if we do not
// change direction, and other factors
module nextBallDir(clk, paddleX, bdir, bx, by, tmpXa, tmpYa, outdir);
input clk;
input [3:0] bdir;
input [9:0] paddleX, bx, tmpXa;
input [8:0] by, tmpYa;
output [3:0] outdir; 

reg [3:0] outdir;

// difference between the leftmost side of the paddle (x-coord) and the ball's x-coord
wire [5:0] diff;
assign diff = bx - paddleX + 30;

always @(posedge clk) begin
  // if we are hitting the paddle, handle separately
  if ((by >= 458) && (by <= 460) && (bx > paddleX - 30) && (bx < paddleX + 30)) begin
    outdir[2] <= 0;
    
    // change direction based on where we hit on the paddle
    if (diff <= 10) begin
      outdir[3] <= 0;
      outdir[1] <= 0;
      outdir[0] <= 1;
    end else if (diff <= 20) begin
      outdir[3] <= 0;
      outdir[1] <= 0;
      outdir[0] <= 0;
    end else if (diff <= 30) begin
      outdir[3] <= 0;
      outdir[1] <= 1;
      outdir[0] <= 0;
    end else if (diff <= 40) begin
      outdir[3] <= 1;
      outdir[1] <= 1;
      outdir[0] <= 0;
    end else if (diff <= 50) begin
      outdir[3] <= 1;
      outdir[1] <= 0;
      outdir[0] <= 0;
    end else begin
      outdir[3] <= 1;
      outdir[1] <= 0;
      outdir[0] <= 1;    
    end
  end else begin
    // not hitting the paddle.  In this case, we will only change x or y direction.  We will NOT change its velocity in any direction
    outdir[1] <= bdir[1];
    outdir[0] <= bdir[0];
    if (!(bx > 2 && bx < 638)) // OB, so change x direction
      outdir[3] <= ~bdir[3];
    // if we are: outside brick region or if we did not cross brick x-lines, don't change direction
    else if (((by < 70) || (by > 300 && by < 478)) || (((tmpXa[5:0] - bx[5:0]) <= 2) || ((bx[5:0] - tmpXa[5:0]) <= 2)))
      outdir[3] <= bdir[3];
    else // we are inside the brick region and we did cross brick x-lines... so we may need to change x-direction (we won't if the brick isn't there)
      outdir[3] <= ~bdir[3];
    
    if (by < 2 || by > 478) // OB, so change Y direction
      outdir[2] <= ~bdir[2];
    else if (!(tmpXa > 2 && tmpXa < 638)) // OB in X direction, so DONT change y direction
      outdir[2] <= bdir[2];
    // if we are: outside brick region or if we did not cross brick y-lines, don't change direction
    else if (((by < 458) || (by > 460) || (bx <= (paddleX - 30)) || (bx >= (paddleX + 30))) && (((by > 2 && by < 70) || (by > 300 && by < 478)) || (((tmpYa[4:0] - by[4:0]) <= 2) || ((by[4:0] - tmpYa[4:0]) <= 2))))
      outdir[2] <= bdir[2];
    else
      outdir[2] <= ~bdir[2];
  end
end

endmodule


